#include "Diffusion.h"
#include "Image.h"

void dNormalizeMaskAlpha(IplImage *image, IplImage *mask) {
    if(!image || !mask) return;
    for(int a = 0; a < image->height; a++) {
        for(int b = 0; b < image->width; b++) {
            CvScalar pixel = cvGet2D( image, a, b );
            double maskValue = 1.0 - (cvGet2D( mask, a, b ).val[0] / 255.0);\
            if(maskValue <= 0.01) continue;
            for(int k = 0; k < image->nChannels; k++) {
                pixel.val[k] /= maskValue ;
                if(pixel.val[k] > 255.0) pixel.val[k] = 255.0;
                if(pixel.val[k] < 0.0) pixel.val[k] = 0.0;
            }
            cvSet2D( image, a, b, pixel);
        }
    }
}

void dPyramidGaussianBlur(IplImage *image, IplImage *mask, int octaves) {
    if(!image || !mask) return;
    if(octaves > MAX_OCTAVES) octaves = MAX_OCTAVES;
    int factorTable[MAX_OCTAVES];
    int factorValue = 2;
    IplImage *output = cvCloneImage( image );
    // Fill out factor table
    for(int a = 0; a < octaves; a++) {
        factorValue = (a * OCTAVE_SCALE) + OCTAVE_OFFSET;
        factorTable[(octaves - 1) - a] = factorValue + 1;
    }
    // Setup - fill in with single median value
    CvScalar avgColor = cvScalar(0.0, 0.0, 0.0, 0.0);
    double numPixels = 0.0f;
    for(int a = 0; a < image->height; a++) {
        for(int b = 0; b < image->width; b++) {
            if( cvGet2D(mask, a, b).val[0] < 128 ) continue;
            CvScalar pixelColor = cvGet2D( image, a, b);
            numPixels++;
            for(int c = 0; c < image->nChannels; c++) {
                avgColor.val[c]+=pixelColor.val[c];
            }
        }
    }
    if(numPixels < 1.0) numPixels = 1.0;
    for(int a = 0; a < image->nChannels; a++) avgColor.val[a] /= numPixels;
    for(int a = 0; a < image->height; a++) {
        for(int b = 0; b < image->width; b++) {
            cvSet2D( output, a, b, avgColor );
            if( cvGet2D(mask, a, b).val[0] > 128 ) 
                cvSet2D( image, a, b, cvScalar(0.0, 0.0, 0.0, 0.0) );
        }
    }
    // Do the pyramid blur work
    IplImage *octImage = NULL;
    IplImage *octMask = NULL;
    for(int a = 0; a < octaves; a++) {
        printl("Diffusing Octave %d, size %d...\n",a , factorTable[a]);
        octImage = cvCloneImage( image );
        octMask = cvCloneImage( mask );
        cvSmooth( octImage, octImage, CV_BLUR, factorTable[a] );
        cvSmooth( octMask, octMask, CV_BLUR, factorTable[a] );
        dNormalizeMaskAlpha( octImage, octMask );
        iMaskedCombine( octImage, output, octMask, false, -1.0 );
        cvReleaseImage( &octImage );
        cvReleaseImage( &octMask );
    }

    iMaskedCombine( output, image, mask, true );

    cvReleaseImage( &output );
}

void dAnisotripicDiffusion( IplImage *image, IplImage *mask, int iter ) {
    IplImage *out = iCloneFloat( image );
    IplImage *dMask = NULL;
    IplImage *coeff = cvCreateImage( cvGetSize( image ), IPL_DEPTH_32F, image->nChannels );
    
    if( mask != NULL) {
        dMask = cvCloneImage( mask );
        cvDilate(dMask, dMask);
        cvSmooth(dMask, dMask);
    }

    vector< pair<int, int> > maskList;

    for(int a = 1; a < image->height - 1; a++) {
        for(int b = 1; b < image->width - 1; b++) {
            if( iPixel(dMask, a, b) > ANISO_MASK_THRES ) {
                maskList.push_back(make_pair(a, b));
            }
        }
    }


    for(int i = 0; i < iter; i++) {

        if(processEvents()) return;
        
        // Calculate coefficients
        for(int j = 0; j < maskList.size(); j++) {
            int a = maskList[j].first;
            int b = maskList[j].second;
            CvScalar cx = cvScalar(0,0,0,0), cy = cvScalar(0,0,0,0), co = cvScalar(0,0,0,0);

            CvScalar pixelx0 = cvGet2D( image, a-1, b+1 ); CvScalar pixelx1 = cvGet2D( image, a  , b+1 );
            CvScalar pixelx2 = cvGet2D( image, a+1, b+1 ); CvScalar pixelx3 = cvGet2D( image, a-1, b-1 );
            CvScalar pixelx4 = cvGet2D( image, a  , b-1 ); CvScalar pixelx5 = cvGet2D( image, a+1, b-1 );

            CvScalar pixely0 = cvGet2D( image, a+1, b-1 ); CvScalar pixely1 = cvGet2D( image, a+1, b   );
            CvScalar pixely2 = cvGet2D( image, a+1, b+1 ); CvScalar pixely3 = cvGet2D( image, a-1, b-1 );
            CvScalar pixely4 = cvGet2D( image, a-1, b   ); CvScalar pixely5 = cvGet2D( image, a-1, b+1 );

            for(int k = 0; k < image->nChannels; k++) {
                cx.val[k] = (pixelx0.val[k] + 2*pixelx1.val[k] + pixelx2.val[k]
                            -pixelx3.val[k] - 2*pixelx4.val[k] - pixelx5.val[k]) * ANISO_COEFF_FACTOR;
                cy.val[k] = (pixely0.val[k] + 2*pixely1.val[k] + pixely2.val[k]
                            -pixely3.val[k] - 2*pixely4.val[k] - pixely5.val[k]) * ANISO_COEFF_FACTOR;
                double q = 1.0 + ( iSq(cx.val[k]) + iSq(cy.val[k] ) );
                co.val[k] = 1.0 / ( 8.0 * sqrt(q) ) / (0.2);
            }

            cvSet2D( coeff, a, b, co );
        }

        // Perform blurring
        for(int j = 0; j < maskList.size(); j++) {
            int a = maskList[j].first;
            int b = maskList[j].second;
            CvScalar o = cvScalar(0,0,0,0);
            CvScalar sum = cvScalar(0,0,0,0);
            for(int a1 = a - 1; a1 <= a + 1; a1++) {
                for(int b1 = b - 1; b1 <= b + 1; b1++) {
                    if( a1 == a && b1 == b ) continue;
                    CvScalar co = cvGet2D( coeff, a1, b1 );
                    CvScalar im = cvGet2D( image, a1, b1 );
                    for(int k = 0; k < image->nChannels; k++) {
                        o.val[k] += co.val[k] * im.val[k];
                        sum.val[k] += co.val[k];
                    }
                }
            }
            CvScalar im_ = cvGet2D( image, a, b );
            for(int k = 0; k < image->nChannels; k++) {
                o.val[k] += (1.0 - sum.val[k]) * im_.val[k] ;
            }
            cvSet2D( out, a, b, o );
         }

         // Convert and output
         double f = 0.0;
         for(int a = 1; a < image->height - 1; a++) {
            for(int b = 1; b < image->width - 1; b++) {
                if(dMask == NULL) f = 0.0; else f = iPixel(dMask, a, b) / 255.0;
                if(mask != NULL && iPixel(mask, a, b) < 128) f = 0.0;
                CvScalar pixel = cvGet2D( out, a, b );
                CvScalar srcpixel = cvGet2D( image, a, b );
                for(int k = 0; k < image->nChannels; k++) {
                    if( pixel.val[k] < 0.0 ) pixel.val[k] = 0.0;
                    if( pixel.val[k] > 255.0 ) pixel.val[k] = 255.0;
                    pixel.val[k] = srcpixel.val[k] * (1.0 - f) + pixel.val[k] * f;
                }
                cvSet2D( image, a, b, pixel );
            }
         }

    }

    cvReleaseImage( &out );
    cvReleaseImage( &dMask );
    cvReleaseImage( &coeff );

}