#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <math.h>
#include <cuda.h>

#include "KernelUtil.cu"

__global__ void gpuDistanceKernel( int2 *dvcSourcePatches, float3 *dvcImage, char *dvcMask, int list,
                                   float *dvcOutput, int patchSize, int px, int py, int height, int width) {
    int i = blockDim.x * blockIdx.x + threadIdx.x;
    if( i >= list ) return;
    int2 dvcSourcePatch = dvcSourcePatches[i];
    float dis = 0.0;
    int a, b;
    for( a = -patchSize; a <= patchSize; a++) {
        for( b = -patchSize; b <= patchSize; b++) {
            int cx1 = dvcSourcePatch.x + a, cy1 = dvcSourcePatch.y + b;
            int cx2 = px + a, cy2 = py + b;
            if( cx1 < 0 || cx1 >= height ) continue;
            if( cy1 < 0 || cy1 >= width ) continue;
            if( cx2 < 0 || cx2 >= height ) continue;
            if( cy2 < 0 || cy2 >= width ) continue;
            char maskPixel = dvcMask[(cx2 * width) + cy2];
            if( maskPixel > 5 ) continue;
            float3 pixel1 = dvcImage[(cx1 * width) + cy1];
            float3 pixel2 = dvcImage[(cx2 * width) + cy2];
            float diffx = pixel1.x - pixel2.x;
            float diffy = pixel1.y - pixel2.y;
            float diffz = pixel1.z - pixel2.z;
            dis += diffx * diffx;
            dis += diffy * diffy;
            dis += diffz * diffz;
        }
    }
    dvcOutput[i] = dis;
}

extern "C" 
float *gpuComputeDistance(int2 *hstSourcePatches, float3 *hstImage, char *hstMask,
                        int height, int width, int list, int patchSize, int px, int py ) {
    
    cudaSetDevice( cutGetMaxGflopsDeviceId() );
    
    const size_t szHstSrcPatches = list * sizeof( int2 );
    const size_t szHstImage = height * width * sizeof( float3 );
    const size_t szHstMask = height * width * sizeof( char );
    const size_t szDvcOutput = list * sizeof( float );
    
    int2 *dvcSourcePatches;
    float3 *dvcImage;
    char *dvcMask;
    float *dvcOutput;

    //printf("CUDA: Allocating memory...\n");
    cudaMalloc( &dvcSourcePatches, szHstSrcPatches );
    cudaMalloc( &dvcImage, szHstImage );
    cudaMalloc( &dvcMask, szHstMask );
    cudaMalloc( &dvcOutput, szDvcOutput );
    if( !dvcSourcePatches || !dvcImage || !dvcMask|| !dvcOutput) return NULL;
    
    //printf("CUDA: Copying memory...\n");
    cudaMemcpy( dvcSourcePatches, hstSourcePatches, szHstSrcPatches, cudaMemcpyHostToDevice ); 
    cudaMemcpy( dvcImage, hstImage, szHstImage, cudaMemcpyHostToDevice ); 
    cudaMemcpy( dvcMask, hstMask, szHstMask, cudaMemcpyHostToDevice ); 
    
    //printf("CUDA: Running kernel threads...\n");
    int numBlocks = (list / 64) + 1;
    int numBlockThreads = min(list, 64);
    gpuDistanceKernel<<< numBlocks, numBlockThreads >>>( dvcSourcePatches,
        dvcImage, dvcMask, list, dvcOutput, patchSize, px, py, height, width );

    //cutilCheckMsg("Kernel execution failed");
    //printf("CUDA: Retrieving results...\n");

    float *output = new float[list];
    if(!output) return NULL;
    cudaMemcpy(output, dvcOutput, szDvcOutput, cudaMemcpyDeviceToHost);

    cudaFree( dvcSourcePatches );
    cudaFree( dvcImage );
    cudaFree( dvcMask );
    cudaFree( dvcOutput );

    return output;
}